from os import environ

# TODO: provide access to modification of these parameters
environ['TRANSFORMERS_OFFLINE'] = "1"
environ['HF_DATASETS_OFFLINE'] = "1"
environ['TF_GPU_ALLOCATOR'] = "cuda_malloc_async"
environ["TOKENIZERS_PARALLELISM"] = "true"

from typing import Union, Iterable, Dict, Any
from pathlib import Path
from datasets import load_dataset #needed

from .cenematic import Cenematic #needed
from .corpus import Corpus #needed


from setuptools import setup

setup(name='semiolog',
      version='0.0.0',
      description='Semiological analysis',
      url='https://gitlab.ethz.ch/semiomaths/semiolog',
      author='Gianni Gastaldi',
      author_email='giannigastaldi@gmail.com',
      license='CC BY-NC 4.0',
      packages=['semiolog'],
      zip_safe=False)


# Chain and Tree are needed elsewhere
from .chain import Chain
from .tree import Tree

from os import path
from tokenizers import (
    decoders,
    models,
    normalizers,
    pre_tokenizers,
    processors,
    Tokenizer,

    Regex,
    NormalizedString,
    PreTokenizedString,
)

from .tokenizer import SequenceSLG, PMI_SLG, NormalizeSLG, Normalize_w_spacesSLG

from transformers import PreTrainedTokenizerFast
import datasets

class Syntagmatic:
    def __init__(self,semiotic) -> None:
        
        self.config = semiotic.config.syntagmatic
        self.config_vocab = semiotic.config.vocabulary
        self.dataset = semiotic.corpus.dataset
        self.split_rate = semiotic.config.corpus.split_rate
        self.path = semiotic.paths.syntagmas
        
        self.tokenizer_path = str(semiotic.paths.vocabulary.joinpath("tokenizer.json"))

        # Load HF Tokenizer model (from "processor" config)

        if self.config.from_file and path.exists(self.tokenizer_path):
            self.tokenizer = Tokenizer.from_file(self.tokenizer_path)
            print(f"SLG [I]: Tokenizer loaded from file")
        else:
            allowed_processors = {"BPE", "WordLevel"}
            if self.config.processor not in allowed_processors:
                raise Exception(f"SLG [E]: Processor provided in config.syntagmatic should be among the following: {allowed_processors}. The processor provided was: {self.config.processor}")
            if self.config.processor == "WordLevel":
                self.tokenizer = Tokenizer(
                    models.WordLevel(
                        vocab = semiotic.vocab.encode,
                        unk_token = self.config_vocab.unk_token
                        )
                        )
            elif self.config.processor == "BPE":
                
                # By the way the vocab is built in SLG, some early merged terms might not appear in the final vocabulary. HF BPE needs them to be there, so they are added here in case it is needed.

                max_vocab_encode = max(semiotic.vocab.encode.values())
                missing_merges = [(k,max_vocab_encode+i+1) for i,k in enumerate({"".join(m) for m in semiotic.vocab.merges}-semiotic.vocab.encode.keys())]

                if missing_merges != []:
                    for k,i in missing_merges:
                        semiotic.vocab.encode[k] = i
                    print(f"SLG [I]: The following merged terms have been added to the vocabulary: {missing_merges}")

                self.tokenizer = Tokenizer(
                    models.BPE(
                        vocab = semiotic.vocab.encode,
                        merges = semiotic.vocab.merges,
                        unk_token = self.config_vocab.unk_token
                        )
                        )
            
            # Load HF normalizer
            # The elif condition on the string SLG is sort of a hack (needed due to non standard declaration of custom normalizer). There should be a more elegant way
            if self.config.normalizer != None:
                if isinstance(self.config.normalizer,list):
                    self.tokenizer.normalizer = normalizers.Sequence(
                        [eval(f"normalizers.{norm}()") for norm in self.config.normalizer]
                        )
                elif self.config.normalizer[-3:] == "SLG":
                    self.tokenizer.normalizer = eval(f"{self.config.normalizer}")
                else:
                    self.tokenizer.normalizer = eval(f"normalizers.{self.config.normalizer}()")

            # Load HF pre-tokenizer
            # The elif condition on the string SLG is sort of a hack (needed due to non standard declaration of custom pre-tokenizer). There should be a more elegant way
            if self.config.pre_tokenizer != None:
                if isinstance(self.config.pre_tokenizer,list):
                    self.tokenizer.pre_tokenizer = pre_tokenizers.Sequence(
                        [eval(f"pre_tokenizer.{pretok}()") for pretok in self.config.pre_tokenizer]
                        )
                elif self.config.pre_tokenizer[-3:] == "SLG":
                    self.tokenizer.pre_tokenizer = eval(f"pre_tokenizers.PreTokenizer.custom({self.config.pre_tokenizer}(semiotic))")
                else:
                    self.tokenizer.pre_tokenizer = eval(f"pre_tokenizers.{self.config.pre_tokenizer}()")
            
            # # Possible post_processor in case needed.

            # cls_token_id = tokenizer.token_to_id(self.config_vocab.cls_token)
            # sep_token_id = tokenizer.token_to_id(self.config_vocab.sep_token)

            # tokenizer.post_processor = processors.TemplateProcessing(
            #     single=f"[CLS]:0 $A:0 [SEP]:0",
            #     pair=f"[CLS]:0 $A:0 [SEP]:0 $B:1 [SEP]:1",
            #     special_tokens=[("[CLS]", cls_token_id), ("[SEP]", sep_token_id)],
            # )

        

        self.bert_tokenizer = PreTrainedTokenizerFast(
            tokenizer_object = self.tokenizer,
            # tokenizer_file=str(semiotic.vocab.path.joinpath("tokenizer.json")), # You can load from the tokenizer file, alternatively

            model_max_length = self.config.model_max_length,
            model_input_names = ["input_ids","token_type_ids", "attention_mask"],

            unk_token= self.config_vocab.unk_token,
            pad_token= self.config_vocab.pad_token,
            cls_token= self.config_vocab.cls_token,
            sep_token= self.config_vocab.sep_token,
            mask_token= self.config_vocab.mask_token,
        )

        # The following manual addition is due to a bug in HF PreTrainedTokenizerFast, not taking into account the special tokens in the previous declaration
        self.special_tokens = {
            "unk_token": self.config_vocab.unk_token,
            "pad_token": self.config_vocab.pad_token,
            "cls_token": self.config_vocab.cls_token,
            "sep_token": self.config_vocab.sep_token,
            "mask_token": self.config_vocab.mask_token,
            }
        self.bert_tokenizer.add_special_tokens(self.special_tokens)

        # The following is a test for debugging. I construct a SLG tokenizer to produce a tokenized string separated by spaces to feed to the main tokenized defined as a simple whitespace splitting

        self.debug_tokenizer = Tokenizer(
                    models.WordLevel(
                        vocab = semiotic.vocab.encode,
                        unk_token = self.config_vocab.unk_token
                        )
                        )
        self.debug_tokenizer.normalizer = NormalizeSLG

        self.debug_tokenizer.pre_tokenizer = pre_tokenizers.PreTokenizer.custom(SequenceSLG(semiotic))

        self.debug_bert_tokenizer = PreTrainedTokenizerFast(
            tokenizer_object = self.debug_tokenizer,
            # tokenizer_file=str(semiotic.vocab.path.joinpath("tokenizer.json")), # You can load from the tokenizer file, alternatively

            model_max_length = self.config.model_max_length,
            model_input_names = ["input_ids","token_type_ids", "attention_mask"],

            unk_token= self.config_vocab.unk_token,
            pad_token= self.config_vocab.pad_token,
            cls_token= self.config_vocab.cls_token,
            sep_token= self.config_vocab.sep_token,
            mask_token= self.config_vocab.mask_token,
        )
        self.debug_bert_tokenizer.add_special_tokens(self.special_tokens)

        # END OF THE DEBUGGING (but see below)


    def save_tokenizer(self):
        self.tokenizer.save(self.tokenizer_path)

    def build(
        self,
        dataset = None,
        save_tokenized = None,
        n_sents = None,
        return_tokenized = False,
        ):

        # TODO: Add save_tokenized and n_sents to config!!!

        if dataset == None:
            dataset = self.dataset

        if save_tokenized == None:
            save_tokenized = self.config.save_tokenized

        if n_sents == None:
            n_sents = self.config.n_sents

        def tokenize_function(syntagmas):

            # DEBUGGING: pre segment string into a string separated by spaces and then feed it to a whitspace tokenizer

            syntagmas = {"text": [" ".join(self.debug_bert_tokenizer.tokenize(sent)) for sent in syntagmas["text"]]}
            
            # END OF DEBUGGING

            return self.bert_tokenizer(syntagmas["text"], return_special_tokens_mask = True)

        print("SLG: Tokenizing dataset...")

        if n_sents !=None:
            dataset = datasets.DatasetDict({
                "train":dataset["train"].select(range(n_sents)),
                "dev": dataset["dev"].select(range(int(n_sents*(1/self.split_rate[0])*self.split_rate[1]))),
                "test": dataset["test"].select(range(int(n_sents*(1/self.split_rate[0])*self.split_rate[2])))
                })


        tokenized_datasets = dataset.map(
            tokenize_function,
            batched = self.config.input_tokenize["batched"],
            batch_size = self.config.input_tokenize["batch_size"],

            # Using more than 1 proc in iMac with m1 blocks the process
            num_proc = 1, #self.cpu_count, 
            remove_columns = self.config.input_tokenize["remove_columns"]
        )
        if save_tokenized:
            tokenized_datasets.save_to_disk(self.path / "tokenized")
            print(f"SLG: Tokenized corpus saved to {self.path / 'tokenized'}")
        
        print("SLG: Corpus tokenized!")

        if return_tokenized:
            return tokenized_datasets




from typing import List
import networkx as nx

from tokenizers import (
    NormalizedString,
    PreTokenizedString,
)

from math import log

from semiolog.util import subsequences


class SequenceSLG:


    """
    """

    def __init__(self, semiotic) -> None:
        self.zipf_factor = .135
        self.semiotic = semiotic

        if self.semiotic.vocab.freq !=None:
            self.voc = self.semiotic.vocab.freq

            # TODO: Zipf factor should (in principle) be computable following Mandelbrot (or not?)
            self.voc_rank = {k:(v+1)**self.zipf_factor for v,k in enumerate(self.voc.keys())}

    def build_graph_data(
        self,
        string: str,
        voc: dict
        )-> List[tuple]:

        edge_data = []
        for beginning in range(0, len(string)):
            for end in range(beginning + 1, len(string) + 1):
                subsequence_label = string[beginning:end]
                if subsequence_label not in voc or subsequence_label == string:
                    continue
                edge_data.append(
                    (
                        beginning,
                        end,
                        {
                            "label": subsequence_label,
                        },
                    )
                )

        return edge_data

    def chain2seq(
        self, string:str
    ) -> List[tuple]:
        lSt = len(string)
        
        # If sent of len <2r, then return the interval
        if lSt<2:
            return [(0,lSt)]
        elif lSt == 2:
            return [(0,1),(1,2)]

        # If a character in the string not in vocab, add it
        for c in string:
            if c not in self.voc:
                self.voc[c]=1
                self.voc_rank[c]=(len(self.voc)+1)**self.zipf_factor

        graph_data = self.build_graph_data(string, self.voc)
        seg_graph_full = nx.DiGraph()
        seg_graph_full.add_edges_from(graph_data)

        # Construct weights
        for edge in seg_graph_full.edges:
            rank = self.voc_rank[seg_graph_full.edges[edge]["label"]]
            seg_graph_full.edges[edge]["weight"] = rank

        # Find best segmentation out of shortest path
        shortest_path = nx.shortest_path(seg_graph_full, 0, lSt, weight="weight")

        seg_offsets = subsequences(shortest_path, 2)

        return seg_offsets

    def SequenceSLG_split(self, i: int, normalized_string: NormalizedString) -> List[NormalizedString]:

        seg_offsets = self.chain2seq(str(normalized_string))

        splits = []
        for start,end in seg_offsets:
            splits.append(normalized_string[start:end])

        return splits

    def pre_tokenize(self, pretok: PreTokenizedString):

        pretok.split(self.SequenceSLG_split)


class PMI_SLG:
    """
    Sequential segmentation maximizing the product of probabilities (which amounts to minizing mutual information between possible components of a given term)
    """

    def __init__(self, semiotic) -> None:
        self.semiotic = semiotic
        self.norm_factor = 1/semiotic.vocab.char_mass
        self.ng_prob = [{k:v*self.norm_factor for k,v in semiotic.vocab.__getattribute__(att).freq.items()} for att in [attribute for attribute in dir(semiotic.vocab) if attribute.startswith("ng")]]
        self.ng_prob = {len(next(iter(ng_i.keys()))):ng_i for ng_i in self.ng_prob}
        self.thres = semiotic.vocab.thres

    def build_graph_data(
        self,
        string: str,
        ng_prob: dict
        )-> List[tuple]:

        edge_data = []
        for beginning in range(0, len(string)):
            for end in range(beginning + 1, len(string) + 1):
                subsequence_label = string[beginning:end]
                if len(subsequence_label) not in ng_prob.keys() or subsequence_label not in ng_prob[len(subsequence_label)] or subsequence_label == string:
                    continue
                edge_data.append(
                    (
                        beginning,
                        end,
                        {
                            "label": subsequence_label,
                        },
                    )
                )

        return edge_data

    def chain2seq(
        self,
        string:str,
        ng_prob:dict,
        ) -> List[tuple]:
        
        lSt = len(string)

        # If sent of len <2r, then return the interval
        if lSt<2:
            return [(0,lSt)]
        elif lSt == 2:
            return [(0,1),(1,2)]

        # If a character in the string not in vocab, add it
        for c in string:
            if c not in ng_prob[1]:
                ng_prob[1][c]=0

        graph_data = self.build_graph_data(string, ng_prob)
        seg_graph_full = nx.DiGraph()
        seg_graph_full.add_edges_from(graph_data)

        # Construct weights

        if lSt not in ng_prob.keys():
            normalizer = 0
        else:
            normalizer = ng_prob[lSt].get(string,0)

        for edge in seg_graph_full.edges:
            label = seg_graph_full.edges[edge]["label"]
            score = ng_prob[len(label)].get(label,0)-normalizer
            
            if score <= self.thres:
                score = float("inf")
            else:
                score = -log(score)

            seg_graph_full.edges[edge]["weight"] = score

        ## Find best segmentation out of shortest path
        # def k_shortest_paths(G, source, target, k, weight=None):
        #     k_s = list(islice(nx.shortest_simple_paths(G, source, target, weight=weight), k))
        #     k_l = [[string[l:r] for l,r in subsequences(sp, 2)] for sp in k_s]
        #     k_sc = [[log(ng_prob[len(label)][label]) for label in k_] for k_ in k_l]
        #     return (k_s,k_l,k_sc)
        # k_short = k_shortest_paths(seg_graph_full, 0, lSt, 5)
        # shortest_path = nx.shortest_path(seg_graph_full, 0, lSt, weight="weight", method="dijkstra")

        shortest_path = nx.shortest_path(seg_graph_full, 0, lSt, weight="weight", method="bellman-ford")

        seg_offsets = subsequences(shortest_path, 2)

        return seg_offsets

    def PMI_SLG_split(self, i: int, normalized_string: NormalizedString) -> List[NormalizedString]:

        seg_offsets = self.chain2seq(str(normalized_string),self.ng_prob)

        splits = []
        for start,end in seg_offsets:
            splits.append(normalized_string[start:end])

        return splits

    def pre_tokenize(self, pretok: PreTokenizedString):

        pretok.split(self.PMI_SLG_split)


import os
from psutil import cpu_count

from .paths import Paths
from .config import Config
from .corpus import Corpus
from .vocabulary import Vocabulary #, nGram
from .syntagmatic import Syntagmatic
from .paradigmatic import Paradigmatic
from .classifier import Classifier
from .typing import Typing
from .text import Text

class Cenematic:
    
    def __init__(
        self,
        name,
        empty=False,
        config_only = False,
        requested_cpu = None,
        vocab = True,
        ) -> None:

        self.name = name
        self.paths = Paths(self.name)
        
        self.config = Config(self)
        if empty == False and os.path.isdir(self.paths.semiotic):
            self.config.from_file()

            # Check config correctness
            print("SLG [I]: Checking config correctness...", end=" ")

            assert sum(self.config.corpus.split_rate) == 1, f"SLG [config.corpus]: Incorrect split rates. Split rates should add to 1 but the ones given add to {sum(self.config.corpus.split_rate)}."

            print("Config correct!")

        self.config.system.cpu_count = cpu_count(logical = False) if requested_cpu == None else requested_cpu

        self.corpus = Corpus(self)
        self.vocab = Vocabulary(self)

        # TODO: configure online repository for models, with automatic download
        
        if not empty:
            
            if os.path.isdir(self.paths.semiotic) and not config_only:
                self.corpus.from_file()
                if vocab:
                    self.vocab.from_file()
            
            else:
                createQ = input(f"No existing model corresponds to the name {self.name}.\n Do you want to create it? (Y/N)")
                if createQ == "Y":
                    os.mkdir(self.paths.semiotic)
                    os.mkdir(self.paths.corpus)
                    os.mkdir(self.paths.corpus / "original")
                    os.mkdir(self.paths.vocabulary)
                    os.mkdir(self.paths.paradigms)
                    self.config.save()
                    print(f"\nFolder for model '{self.name}' created in {self.paths.models.absolute()}")
                
        self.syntagmatic = Syntagmatic(self)
        # self.ng2 = nGram()
        self.paradigmatic = Paradigmatic(self)
        # self.classifier = Classifier(self)
        # self.typing = Typing(self)
        
        # # Load universal dependencies (ud) and constituency parsing (cp) models
        # self.ud = spacy.load(self.config["evaluation"]["ud_model"])
        # self.cp = spacy.load(self.config["evaluation"]["cp_model"])
        # self.cp.add_pipe("benepar", config={"model": "benepar_en3"})

    def __repr__(self) -> str:
        return f"Cenematic({self.name})"

    def __call__(self,input_chain,paradigms=True):
        return Text(input_chain,self,paradigms=paradigms)